import torch


class SelfAttention(torch.nn.Module):
    def __init__(self, embed_dim):
        super().__init__()

        self.embed_dim = embed_dim
        # TODO: Create linear layers for generating query, key, and value vectors of size embed_dim
        # embed_dim is both the dimension of the input vector and the Q, K, V vectors
        self.query_projection = None
        self.key_projection = None
        self.value_projection = None
        # TODO: Optional: initialise a softmax layer here.
        # Otherwise, you may also use torch.nn.functional.softmax directly in the forward function
        self.softmax = None

    def forward(self, x):
        # TODO: Compute/generate queries, keys and values
        # Step 1: Pass the input through the projection layers to get queries, keys and values
        queries = None
        keys = None
        values = None
        # TODO: Compute the scores (SCALED dot products) as described below
        # Step 2: Get the scores, i.e., the dot products between queries and keys
        # queries, keys and values have shape [batch_size, sequence_length, embed_dim]
        # The matrix containing the scores between all pairs of queries and keys
        # should have shape [batch_size, sequence_length, sequence_length]
        # where matrix[batch_index, i, j] = dot_product(q[batch_index, i], k[batch_index, j]) / sqrt(embed_dim)
        # This can be achieved using torch.matmul or torch.bmm (bmm = batch matrix multiply) instead of loops.
        # Think how can one of the 2 matrices can be transformed to get the desired result.
        # Hint: multiplying matrix of shape [a, b] with [b, c] results in a matrix of shape [a, c]
        #       Use .transpose() function of the tensor to swap two dimensions
        scores = None
        # TODO: Convert the scores into attention weights
        # Step 3: Get the attention weights
        # Run the matrix either through the softmax layer created in __init__ or use torch.nn.functional.softmax
        # Don't forget to specify the correct dim for softmax.
        attention_weights = None
        # TODO: Compute and return the context vectors
        # Step 4: Weighted average of value vectors
        # Finally, return the weighted average of the value vectors.
        # You can use torch.matmul or torch.bmm again here.
        return None


class MultiHeadSelfAttention(torch.nn.Module):
    def __init__(self, embed_dim, num_heads):
        super().__init__()
        assert embed_dim % num_heads == 0, 'Embedding dimension must be a multiple of number of heads.'

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        # TODO: Create linear layers for generating query, key, and value vectors of size embed_dim
        self.query_projection = None
        self.key_projection = None
        self.value_projection = None
        self.output_projection = None
        # TODO: Optional: initialise a softmax layer here.
        # Otherwise, you may also use torch.nn.functional.softmax directly in the forward function
        self.softmax = None

    def forward(self, x):
        # TODO: Compute/generate queries, keys and values
        # Step 1: Pass the input through the projection layers to get queries, keys and values
        queries = None
        keys = None
        values = None
        # TODO: Split Q, K, V vectors into num_heads to get separate Q, K, V matrices for each head
        # Step 2: Split the queries, keys and values into num_heads to
        # have shape [batch, seq_length, num_heads, embed_dim // num_heads]
        # Hint: You just need to use the .reshape() function
        queries = None
        keys = None
        values = None
        # TODO: Generate/compute the scores using queries and keys
        # Step 3: Compute the scores
        # First, permute Q, K to have shape [batch, num_heads, seq_length, embed_dim // num_heads]
        queries = None
        keys = None
        # TODO: Compute the SCALED dot products
        # Now do the matrix multiplication between queries and keys per head to get scores.
        # Hint: You don't need loops. Use torch.matmul instead. And forget to swap dimensions before matmul.
        scores = None
        # TODO: Convert the scores into attention weights
        # Step 4: Attention weights
        # Run the matrix either through the softmax layer created in __init__ or use torch.nn.functional.softmax
        attention_weights = None
        # TODO: Compute the context vectors
        # Step 5: Weighted average
        # Permute values to have shape [batch, num_heads, seq_length, embed_dim // num_heads] just like above
        values = None
        # Now compute the weighted average
        context_vectors = None
        # Concatenate the outputs of different heads
        context_vectors = None  # Hint: Permute the dims and then reshape the tensor
        # TODO: Do the final linear projection and return the output
        # Step 6: Run the output through the final linear layer
        return None


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


if __name__ == '__main__':
    batch_size = 128
    seq_len = 100
    embedding_len = 512
    multi_head_att = MultiHeadSelfAttention(embedding_len, 8)
    att = SelfAttention(embedding_len)
    input_x = torch.randn(batch_size, seq_len, embedding_len)
    print(f'Multi-headed self-attention has {count_parameters(multi_head_att)} parameters.')
    print(f'Single-headed self-attention has {count_parameters(att)} parameters.')
    print(f'Output of multi-headed self-attention is of shape {multi_head_att(input_x).shape}')
    print(f'Output of single-headed self-attention is of shape {att(input_x).shape}')
